# flake8: noqa
FORMAT_INSTRUCTIONS = """You are an AI system that only responds
with a single complete Thought, Action, Action Input format
OR a single Final Answer format.

Complete Format:

Thought: (reflect on your progress and decide what to do next)
Action: (the action name, should be one of [{tool_names}])
Action Input: (the input string to the action) 

OR

Final Answer: (the final answer to the original input question)

"""
QUESTION_PROMPT = """
Answer the question below using the following tools:

{tool_strings}

Question: {input}
"""
SUFFIX = """
Thought: {agent_scratchpad}
"""
FINAL_ANSWER_ACTION = "Final Answer:"



from typing import Dict, List, Tuple, Union

from langchain.agents import AgentExecutor
from langchain.schema import AgentAction, AgentFinish, OutputParserException
from langchain.tools import BaseTool


class ExceptionTool(BaseTool):
    name = "_Exception"
    description = "Exception tool"

    def _run(self, query: str) -> str:
        return query

    async def _arun(self, query: str) -> str:
        return query


class RetryAgentExecutor(AgentExecutor):
    """Agent executor that retries on output parser exceptions."""

    def _take_next_step(
        self,
        name_to_tool_map: Dict[str, BaseTool],
        color_mapping: Dict[str, str],
        inputs: Dict[str, str],
        intermediate_steps: List[Tuple[AgentAction, str]],
    ) -> Union[AgentFinish, List[Tuple[AgentAction, str]]]:
        try:
            return super()._take_next_step(
                name_to_tool_map, color_mapping, inputs, intermediate_steps
            )
        except OutputParserException as e:
            # ok, this isn't great - I agree
            # but not sure of a more clean way to get this
            text = str(e).split("`")[1]
            observation = "Invalid or incomplete response"
            agent_action = AgentAction("_Exception", observation, text)
            self.callback_manager.on_agent_action(
                agent_action, verbose=self.verbose, color="red"
            )
            tool_run_kwargs = self.agent.tool_run_logging_kwargs()
            # trigger run to get bookkeeping on basetool callback manager
            ExceptionTool().run(
                observation, verbose=self.verbose, color="red", **tool_run_kwargs
            )
            return [(agent_action, observation)]



from __future__ import annotations
from langchain.chat_models.base import BaseChatModel
from langchain.prompts import PromptTemplate
from langchain.chains import LLMChain
from langchain.callbacks.base import BaseCallbackManager
from .prompts import FORMAT_INSTRUCTIONS, SUFFIX, QUESTION_PROMPT
from langchain.agents.agent import Agent, AgentOutputParser
from typing import Any, Optional, Sequence
from langchain.tools import BaseTool
from langchain.agents.mrkl.base import ZeroShotAgent
from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
    AIMessagePromptTemplate,
)


class ChatZeroShotAgent(ZeroShotAgent):
    """Agent for the MRKL chain."""

    @classmethod
    def create_prompt(
        cls,
        tools: Sequence[BaseTool],
        suffix: str = SUFFIX,
        format_instructions: str = FORMAT_INSTRUCTIONS,
    ) -> PromptTemplate:
        """Create prompt in the style of the zero shot agent.

        Args:
            tools: List of tools the agent will have access to, used to format the
                prompt.
            prefix: String to put before the list of tools.
            suffix: String to put after the list of tools.
            input_variables: List of input variables the final prompt will expect.

        Returns:
            A PromptTemplate with the template assembled from the pieces here.
        """
        tool_strings = "\n".join(
            [f"    {tool.name}: {tool.description}" for tool in tools]
        )
        tool_names = ", ".join([tool.name for tool in tools])
        format_instructions = format_instructions.format(
            tool_names=tool_names, tool_strings=tool_strings
        )
        human_prompt = PromptTemplate(
            template=QUESTION_PROMPT,
            input_variables=["input"],
            partial_variables={"tool_strings": tool_strings},
        )
        human_message_prompt = HumanMessagePromptTemplate(prompt=human_prompt)
        ai_message_prompt = AIMessagePromptTemplate.from_template(suffix)
        system_message_prompt = SystemMessagePromptTemplate.from_template(
            format_instructions
        )
        # ignore suffix
        return ChatPromptTemplate.from_messages(
            [system_message_prompt, human_message_prompt, ai_message_prompt]
        )

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: BaseChatModel,
        tools: Sequence[BaseTool],
        callback_manager: Optional[BaseCallbackManager] = None,
        output_parser: Optional[AgentOutputParser] = None,
        suffix: str = SUFFIX,
        format_instructions: str = FORMAT_INSTRUCTIONS,
        **kwargs: Any,
    ) -> Agent:
        """Construct an agent from an LLM and tools."""
        cls._validate_tools(tools)
        prompt = cls.create_prompt(
            tools,
            suffix=suffix,
            format_instructions=format_instructions,
        )
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
            callback_manager=callback_manager,
        )
        tool_names = [tool.name for tool in tools]
        _output_parser = output_parser or cls._get_default_output_parser()
        return cls(
            llm_chain=llm_chain,
            allowed_tools=tool_names,
            output_parser=_output_parser,
            **kwargs,
        )



from rmrkl import ChatZeroShotAgent, RetryAgentExecutor
from langchain.llms.fake import FakeListLLM
from langchain.agents import load_tools


def test_agent_init():
    tools = load_tools(["python_repl"])
    responses = [
        "I should use the REPL tool",
        "Action: Python REPL\nAction Input: print(2 + 2)",
        "Final Answer: 4",
    ]
    llm = FakeListLLM(responses=responses)

    agent = RetryAgentExecutor.from_agent_and_tools(
        tools=tools,
        agent=ChatZeroShotAgent.from_llm_and_tools(llm, tools),
        verbose=True,
    )
    agent.run("What is 2 + 2")



from setuptools import setup

exec(open("rmrkl/version.py").read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="rmrkl",
    version=__version__,
    description="Robust Modular Reasoning, Knowledge and Language agent that uses tools and retries on failure",
    author="Andrew White",
    author_email="andrew.white@rochester.edu",
    url="https://github.com/whitead/robust-mrkl",
    license="MIT",
    packages=["rmrkl"],
    install_requires=["langchain>=0.0.147"],
    test_suite="tests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
)
